.isqx-symbol-where,
.isqx-symbol-remarks {
    opacity: .5;
}

.isqx {
    display: flex;
    flex-direction: column;
    row-gap: .5rem;
}

.isqx-wikidata,
.isqx-symbols,
.isqx-detail {
    border-left: 3px solid #88888822;
    padding-left: .5rem;
    border-radius: 4px;
}

.isqx-detail .isqx-where {
    display: grid;
    grid-template-columns: max-content max-content 1fr;
    align-items: baseline;
    gap: .25em .5em;
}

.isqx-detail .isqx-where-row {
    display: contents;
}

.isqx-detail .isqx-symbol {
    text-align: right;
}

.isqx-highlight {
    border-radius: 4px;
    background-color: rgba(var(--highlight-rgb), .1);
    box-shadow: 0 0 6px rgba(var(--highlight-rgb), .3);
}

.isqx-highlight.isqx-hover {
    background-color: rgba(var(--highlight-rgb), .4);
    box-shadow: 0 0 6px rgba(var(--highlight-rgb), .8);
}