{% extends "_base/attribute.html.jinja" %}
{% import "isqx/isqx.jinja" as isqx %}

{% block contents scoped %}
{{ super() }}
<div class="isqx">
    {{ isqx.show_wikidata(attribute) }}
    {{ isqx.show_symbols(attribute) }}
    {{ isqx.show_details(attribute) }}
</div>
{% endblock contents %}