"""Citations for `isqx`.

They can be used for cross-references in the documentation, e.g.
`see [special publication 811][isqx._citations.SP811]`.
A tiny griffe extension is used to copy the runtime value into the docstring.
"""

SI = r"""```bibtex
@techreport{si,
    author = {{Bureau International des Poids et Mesures}},
    title = {The International System of Units ({SI})},
    edition = {9th},
    year = {2019},
    note = {Version 3.01, updated August 2024},
    institution = {Bureau International des Poids et Mesures (BIPM)},
    address = {Sèvres, France},
    url = {https://www.bipm.org/documents/20126/41483022/SI-Brochure-9-EN.pdf}
}
```"""
IUPAP1 = r"""```bibtex
@techreport{iupap1,
    author = {{IUPAP Commission C2 on Symbols, Units, Nomenclature, Atomic Masses and Fundamental Constants}},
    title = {Symbols, Units, Nomenclature and Fundamental Constants in Physics},
    year = {1987},
    institution = {International Union of Pure and Applied Physics (IUPAP)},
    url = {https://archive2.iupap.org/wp-content/uploads/2014/05/A4.pdf}
}
```"""
SP811 = r"""```bibtex
@techreport{sp811,
    title = {Guide for the Use of the International System of Units ({SI})},
    year = {2008},
    edition = {2008},
    institution = {National Institute of Standards and Technology},
    type = {Special Publication},
    number = {811},
    address = {Gaithersburg, MD},
    url = {https://www.nist.gov/pml/special-publication-811}
}
```"""
CODATA2022 = r"""```bibtex
@techreport{codata2022,
    title = {{CODATA} Recommended Values of the Fundamental Physical Constants: 2022},
    year = {2024},
    month = {May},
    institution = {National Institute of Standards and Technology},
    type = {Special Publication},
    number = {961},
    url = {https://physics.nist.gov/cuu/pdf/sp961.pdf}
}
```"""
ISO_80000_2 = r"""```bibtex
@standard{iso_80000_2,
    organization = {International Organization for Standardization},
    title = {Quantities and units - Part 2: Mathematics},
    year = {2019},
    number = {ISO 80000-2:2019},
    url = {https://www.iso.org/standard/76921.html}
}
```"""
ISO_80000_3 = r"""```bibtex
@standard{iso_80000_3,
    organization = {International Organization for Standardization},
    title = {Quantities and units - Part 3: Space and time},
    year = {2019},
    number = {ISO 80000-3:2019},
    url = {https://www.iso.org/standard/64974.html}
}
```"""
ISO_80000_4 = r"""```bibtex
@standard{iso_80000_4,
    organization = {International Organization for Standardization},
    title = {Quantities and units - Part 4: Mechanics},
    year = {2019},
    number = {ISO 80000-4:2019},
    url = {https://www.iso.org/standard/64975.html}
}
```"""
ISO_80000_5 = r"""```bibtex
@standard{iso_80000_5,
    organization = {International Organization for Standardization},
    title = {Quantities and units - Part 5: Thermodynamics},
    year = {2019},
    number = {ISO 80000-5:2019},
    url = {https://www.iso.org/standard/64976.html}
}
```"""
IEC_80000_6 = r"""```bibtex
@standard{iec_80000_6,
    organization = {International Electrotechnical Commission},
    title = {Quantities and units - Part 6: Electromagnetism},
    year = {2022},
    number = {IEC 80000-6:2022},
    url = {https://www.iso.org/standard/77846.html}
}
```"""
IEV = r"""```bibtex
@misc{iev,
    author = {{IEC}},
    title = {Electropedia - International Electrotechnical Vocabulary},
    organization = {International Electrotechnical Commission},
    url = {https://www.electropedia.org/}
}
```"""
ISO_80000_7 = r"""```bibtex
@standard{iso_80000_7,
    organization = {International Organization for Standardization},
    title = {Quantities and units - Part 7: Light and radiation},
    year = {2019},
    number = {ISO 80000-7:2019},
    url = {https://www.iso.org/standard/64977.html}
}
```"""
ISO_80000_8 = r"""```bibtex
@standard{iso_80000_8,
    organization = {International Organization for Standardization},
    title = {Quantities and units - Part 8: Acoustics},
    year = {2020},
    number = {ISO 80000-8:2020},
    url = {https://www.iso.org/standard/64978.html}
}
```"""
ISO_80000_9 = r"""```bibtex
@standard{iso_80000_9,
    organization = {International Organization for Standardization},
    title = {Quantities and units - Part 9: Physical chemistry and molecular physics},
    year = {2019},
    number = {ISO 80000-9:2019},
    url = {https://www.iso.org/standard/64979.html}
}
```"""
ISO_80000_10 = r"""```bibtex
@standard{iso_80000_10,
    organization = {International Organization for Standardization},
    title = {Quantities and units - Part 10: Atomic and nuclear physics},
    year = {2019},
    number = {ISO 80000-10:2019},
    url = {https://www.iso.org/standard/64980.html}
}
```"""
ISO_80000_11 = r"""```bibtex
@standard{iso_80000_11,
    organization = {International Organization for Standardization},
    title = {Quantities and units - Part 11: Characteristic numbers},
    year = {2019},
    number = {ISO 80000-11:2019},
    url = {https://www.iso.org/standard/64982.html}
}
```"""
ISO_80000_12 = r"""```bibtex
@standard{iso_80000_12,
    organization = {International Organization for Standardization},
    title = {Quantities and units - Part 12: Solid state physics},
    year = {2019},
    number = {ISO 80000-12:2019},
    url = {https://www.iso.org/standard/63480.html}
}
```"""
ISO_80000_13 = r"""```bibtex
@standard{iso_80000_13,
    organization = {International Organization for Standardization},
    title = {Quantities and units - Part 13: Information science and technology},
    year = {2025},
    number = {ISO 80000-13},
    url = {https://www.iso.org/standard/87648.html}
}
```"""
MP_UNITS = r"""```bibtex
@software{mp_units,
    author = {Pusz, Mateusz and Guerrero Peña, Johel Ernesto and Hogg, Chip and {The mp-units project team}},
    license = {MIT},
    month = nov,
    title = {{mp-units}},
    url = {https://github.com/mpusz/mp-units},
    version = {2.4.0},
    year = {2024}
}
```"""
# usc
H44 = r"""```bibtex
@techreport{h44,
    author = {{NIST}},
    title = {NIST Handbook 44 - 2024 - Appendix C. General Tables of Units of Measurement},
    year = {2024},
    institution = {National Institute of Standards and Technology},
    url = {https://www.nist.gov/document/nist-handbook-44-2024-appendix-c-pdf}
}
```"""
WMA1985 = r"""```bibtex
@misc{wma1985,
    title = {Weights and Measures Act 1985},
    year = {1985},
    organization = {{legislation.gov.uk}},
    url = {https://www.legislation.gov.uk/ukpga/1985/72/contents}
}
```"""
# aerospace
ICAO = r"""```bibtex
@misc{icao,
    author = {{ICAO}},
    title = {Annex 5 - Units of Measurement to be Used in the Air and Ground Services},
    organization = {International Civil Aviation Organization},
    url = {https://store.icao.int/en/annex-5-units-of-measurement-to-be-used-in-the-air-and-ground-services}
}
```"""
