.panel {
  display: flex;
  flex-direction: column;
  gap: 0.5rem;
}

.controls {
  display: flex;
  gap: 0.75rem;
  padding-bottom: 0.75rem;

  button {
    background: var(--bg-button);
    color: var(--text-primary);
    cursor: pointer;
    border-radius: 6px;
    padding: 0.25rem 0.5rem;
    font-size: 14px;
    border: none;
    &:hover {
      background: var(--bg-button-hover);
    }
  }
}

.nodeDetails {
  display: flex;
  flex-direction: column;
  gap: 0.5rem;
  padding-bottom: 1rem;
}

.nodeDetail,
.definedInDetail {
  border-radius: 5px;
  padding: 0.25rem 0.25rem 0.25rem 0.75rem;
  background-color: #24242888;
  border: 1px solid #323237;
  border-left-width: 4px;
  border-left-style: solid;
  transition: border-left-color 0.2s ease-in-out;

  .nodeHeader {
    display: flex;
    align-items: baseline;
    gap: 0.5rem;
    flex-wrap: wrap;

    h3 {
      margin: 0;
      font-size: 1.3rem;
      user-select: none;
      cursor: pointer;
      font-weight: 700;
    }
  }

  .symbols {
    font-size: 1.1em;
  }
}

.separator {
  margin: 0.25rem 0;
  border: 0;
  border-top: 1px solid var(--bg-border);
}

.detailSection {
  margin-top: 0.5rem;
  margin-bottom: 0.5rem;
  padding-right: 0.5rem;

  h4 {
    margin: 0 0 0.5rem 0;
    border-bottom: 1px solid var(--bg-border);
    padding-bottom: 0.25rem;
    font-size: 1rem;
  }

  .definedIn {
    display: flex;
    flex-direction: column;
    gap: 0.5rem;
  }

  a {
    color: var(--accent-link, #a5b8c4);
    text-decoration: none;
    &:hover {
      text-decoration: underline;
    }
  }
}

.equationBlock {
  overflow-x: auto;

  &::-webkit-scrollbar {
    display: none;
  }
}

.whereClause {
  border-radius: 5px;
  border-left: 4px solid var(--bg-border);
  padding-left: 0.5rem;
  font-size: 0.9em;
  display: grid;
  grid-template-columns: max-content max-content 1fr;
  align-items: baseline;
  gap: 0.25rem 0.25rem;
  padding-top: .25rem;
  padding-bottom: .25rem;

  .whereRow {
    display: contents;
  }

  .whereSymbol {
    text-align: right;
  }
}



.symbolRemarks {
  color: var(--text-muted);
}